#
# Provides functions for executing pi-expressions in asynchronous pi-calculus
# Copyright (C) 2007 Frank Puhlmann.
#

require 'pimodel.rb'

class AsyncPiExecutor

	# Required to detect recursive agent definitions for validation
	validatedAgents = Array.new
	# Set of restricted names
	restrictions = Array.new	
	# Counts the number of restrictions for each restricted name
	restrictionCounter = Hash.new
	restrictionCounter.default=0
	# Counts the execution steps for updating the identifier
	executionCounter = 1
	# Keeps the original agent
	
	def initialize(agent,*validate)
	
		if agent.class != AgentRef
			raise "AgentRef required!"
			end
	
		# Set up class variables
		@validatedAgents = Array.new
		@restrictionCounter = Hash.new
		@restrictionCounter.default=-1
		@executionCounter = 0
		@restrictions = Array.new
	
		if validate.size==0
			# Add agent to already validated
			@validatedAgents << agent.agent.behavior
			# Check agent for syntactical correctness
			validate(agent)
			end
	
		# Create a copy of the agent
		@currentAgent = flatCopy(agent)
		@origAgent = agent
		
	end
	
	def currentAgent
		agent = Agent.new(@origAgent.agent.ident,@origAgent.agent.params,@currentAgent)
		agent.restrictions = @restrictions.clone
		# Update identifier
		agent.ident = agent.ident.sub(/\'[0-9]*/, "")
	    agent.ident = agent.ident + "'" + @executionCounter.to_s if @executionCounter > 0
	    # Optimize restrictions
		optimizeRestrictions(agent)
		return agent		
	end
	
	# Validates if the agent confirms to the asyncronous pi-calculus syntax
	def validate(beh)
			
		# Check if behavior exists
		if !beh
			return
		end
		
		#puts "VALIDATING " + beh.to_s + " (" + beh.class.to_s + ")"
		
		case beh
		
			when Composition:
				# Allowed parts of a composition node is anything, so just get on...
				beh.parts.each { |p| validate(p) }
				
			when Summation:
				# Only input and tau in a summation is allowed
				beh.parts.each do |p|
					case p
						when Input, Tau, Summation:
							validate(p)
						else
							raise "VALIDATION ERROR: INVALID SUMMAND FOUND IN " + beh.to_s + " (" + p.class.to_s + ")"
					end
				end
				
			when Output:
				# An output must not have a successor
				if beh.successor
					raise "VALIDATION ERROR: OUTPUT WITH SUCCESSOR FOUND IN " + beh.to_s
				end
				
			when Input, Tau:
				# An input or tau might have a successor
				if beh.successor 
					validate(beh.successor)
				end
				
			when Restriction:
				# A restriction must have a successor
				if !beh.successor
					raise "VALIDATION ERROR: RESTRICTION WITHOUT SUCCESSOR FOUND IN " + beh.to_s
				end
				validate(beh.successor)
				
			when AgentRef:
				# Jump into (if not already checked)
				if !@validatedAgents.include?(beh)
					# Add
					@validatedAgents << beh
					# Check
					validate(beh.behavior) if beh.instance_of? Agent
					validate(beh.agent.behavior) if beh.instance_of? AgentRef
				end
				
			when Agent:
				raise "VALIDATION ERROR: AGENT FOUND (AGENTREF EXPECTED) IN " + beh.to_s
			
			else
				# Raise error
				raise "VALIDATION ERROR: UNKNOWN NODE TYPE: " + beh.class.to_s
		end
		
	end

	# Executes a certain node of an Agent
	def exec(node)
	    
	   case node
	   
	   		when Node:
	   		
				p = node.parent
				
				case p.value
				
				when Summation:
					# Remove complete summation 
					# (until Composition or nil is found, according to SC-SUM-ASSOC)
					c = node
					while (p.value.instance_of? Summation) 
						c = p
						p = p.parent
						end 
						
					case p.value
					
						when Composition:		
							# Add entry in composition, if composition
							index = p.value.parts.index(c.value)
							#puts "INDEX = " + index.to_s
							if node.value.successor
								# Add successor 
								p.value.parts[index]=flatCopy(node.value.successor)
							else
								# Remove entry				
								p.value.parts.delete_at(index)
							end
					
						when nil:
							# Add entry as agent behavior
							if node.value.successor
								currentAgent.behavior = flatCopy(node.value.successor)
							else 
								currentAgent.behavior = nil
							end
						
						else
						
						puts "ERR: UNSUPPORTED NODE TYPE: " + p.value
						
					end
				
				when Composition:
					index = p.value.parts.index(node.value)
					if node.value.successor
						# Add successor 				
						p.value.parts[index]=flatCopy(node.value.successor)
					else
						# Remove entry				
						p.value.parts.delete_at(index)
					end
					cleanUp(p)
						
				when nil:
				
					# Simply move on to successor
					# Test for expansion of AgentRef
					#_expandAgentRef(node.value)					
					currentAgent.behavior = flatCopy(node.value.successor)
						
				end
				
			when IntraactionNodes:
				
					# Send must be in composition
					p1 = node.output.parent
					case p1.value
						when Composition:
							# Remove entry
							index = p1.value.parts.index(node.output.value)
							#puts "DELETING " + node.output.value.to_s + " at index " + index.to_s 
							p1.value.parts.delete_at(index) if index
							cleanUp(p1)
						else
							raise "ERR: ILLEGAL PARENT FOR OUTPUT PREFIX: " + node.output.to_s
					end
					
					# Receive can be in composition or summation!
					p2 = node.input.parent
					case p2.value
						when Composition:
							index = p2.value.parts.index(node.input.value)
							if node.input.value.successor					
								# Add successor 
								p2.value.parts[index]=flatCopy(node.input.value.successor)
								# Apply substitution
								if node.output.value.subjects.size != node.input.value.subjects.size
									raise "ERR: SUBJECT SIZE MISMATCH IN " + node.to_s
								end
								
								node.output.value.subjects.size.times do |i|
									alphaConvert(p2.value.parts[index],
									 			 node.input.value.subjects[i],
									 			 node.output.value.subjects[i])
								end
								
							else
								# Remove entry				
								p2.value.parts.delete_at(index)
								cleanUp(p2)
							end
							
						
						when Summation:
							#Remove complete summation 
							#(until Composition or nil is found, according to SC-SUM-ASSOC)
							c = node
							#p2 = p2.input.parent
							while (p2.value.instance_of? Summation) 
								c = p2
								p2 = p2.parent
								end 
						
							case p2.value
						
								when Composition:		
									index = p2.value.parts.index(c.value)	
									# Add entry in composition, if composition
									if node.input.value.successor		
										# Add successor 
										p2.value.parts[index]=flatCopy(node.input.value.successor)
										# Apply substitution
								if node.output.value.subjects.size != node.input.value.subjects.size
									raise "ERR: SUBJECT SIZE MISMATCH IN " + node.to_s
								end
								
										node.output.value.subjects.size.times do |i|
											alphaConvert(p2.value.parts[index],
									 					 node.input.value.subjects[i],
										 				 node.output.value.subjects[i])
										end									
									else
										# Remove entry				
										p2.value.parts.delete_at(index) if index
										cleanUp(p2)
									end
							
								when nil:
									# Add entry as agent behavior
									if node.input.value.successor
										currentAgent.behavior = flatCopy(node.value.successor)
										# Apply substitution
								if node.output.value.subjects.size != node.input.value.subjects.size
									raise "ERR: SUBJECT SIZE MISMATCH IN " + node.to_s
								end
								
										node.output.value.subjects.size.times do |i|
											alphaConvert(p2.value.parts[index],
									 					 node.input.value.subjects[i],
										 				 node.output.value.subjects[i])
										end											
									else 
										currentAgent.behavior = nil
									end
	
	
						end
						
						else
	   						raise "ERR: ILLEGAL PARENT FOR INPUT PREFIX: " + node.input.to_s
					end
		
	   		else
	   
	   			puts "ERR: UNSUPPORTED TRANSITION!"
	   			return
	   			
	   end
	   
	   # Optimize restrictions
	   optimizeRestrictions(currentAgent)
	   
	   # Change identifier
	   @executionCounter = @executionCounter + 1
	
	end

		
	# Collect all active prefixes with backward linkage (up to defined agents)
	def collectActivePrefixes
	
		node = currentAgent.behavior

		active_prefixes = Array.new
		
		# Test, if agent is not empty
		if !node 
			return active_prefixes	
		end
		
		# Collect "front-standing" prefixes 
		_collectActivePrefixes(node,Node.new(currentAgent,nil),active_prefixes)
		
		# Create lists of input and output prefixes
		inputPrefixes, outputPrefixes, tauPrefixes, matchNodes = Array.new, Array.new, Array.new, Array.new
		active_prefixes.each do |n| 
		
			#puts "LIST: " + n.value.to_s + " / " + n.value.class.to_s if n
		
			case n.value
			
			when Input:
				inputPrefixes << n 
		
			when Output:		
				outputPrefixes << n
				
			when Tau:
				tauPrefixes << n
				
			end if n
		end if active_prefixes
		
#		puts "Current Input-prefixes: "
#		inputPrefixes.each { |i| puts i.value.to_s }
#			
#		puts "Current Output-prefixes: "
#		outputPrefixes.each { |i| puts i.value.to_s }
#		
#		puts "Current Tau-prefixes: "
#		tauPrefixes.each { |i| puts i.value.to_s }
		
		# Match input and output prefixes
		outputPrefixes.each do |out|
				
			# Test, if output is in a composition relation (REQUIRED BY WELL-FORMEDNESS!)
			if !(out.parent.value.instance_of? Composition)
				puts "WARNING: OUTPUT IN SUMMATION IGNORED: " + out.value.to_s + 
				     " in " + out.parent.value.to_s	
				next
			end	
			
			inputPrefixes.each do |inp|
			
				#puts "TEST: " + out.value.object.to_s + " vs. " + inp.value.object.to_s
			
				if out.value.object.eql?(inp.value.object)
	
					matchNodes << IntraactionNodes.new(out,inp) 
				
				end
				
			end
			
		end
		
		return (tauPrefixes << matchNodes).flatten!
	end
		
private
	
	def _collectActivePrefixes(node,container,result)
	
		case node
		
		when Restriction:	
			# Continue behind restriction
			#result << Node.new(node.successor,_collectActivePrefixes(node.successor,node,result)) if node.successor
			_collectActivePrefixes(node.successor,node,result) if node.successor
			result.flatten!
		
		when Composition, Summation:	
		    node.parts.each do |c| 
		 		_collectActivePrefixes(c,Node.new(container,node),result) 
		 		end if node.parts
			 	#puts result.inspect
	        result.flatten!
	         
	    when Input, Output, Tau:
	    
			result << Node.new(container, node)
			
		else		
			raise "WHOOPS, unknown node type: " + node.class.to_s
			
		end
	
	end


	
	#
	# Resolves restrictions in an agent by substituting restricted names with unique ones
	#
	def alphaConvert(beh,from,to)
	
		#puts "Alpha-converting " + from + " to " + to + " in " + beh.to_s + " (" + beh.class.to_s + ")"

		# Iterate of beh
		case beh
		
			when Restriction:
				# Stop if from is restricted (again)
				if beh.name.eql?(from) 
					return
				else 
					alphaConvert(beh.successor,from,to) if beh.successor
				end
		
			when Composition, Summation:
				beh.parts.each do |c| 
					alphaConvert(c,from,to) 
					end
				
			when Input:
							
				beh.object=to if beh.object.eql?(from) 
				# Check if from is not in subjects
				beh.subjects.each do |o| 
					if o.eql?(from) 
						return 
					end  
				end if beh.subjects
				alphaConvert(beh.successor,from,to) if beh.successor
				
			when Output:
					
				beh.object=to if beh.object.eql?(from) 
				beh.subjects.size.times do |i| 
					if beh.subjects[i].eql?(from) 
						beh.subjects[i]=to  
					end
				end if beh.subjects
				
			when Tau:
				alphaConvert(beh.successor,from,to) if beh.successor
				
			when AgentRef:
			
				#puts "AGENTREF IN ALPHA"
			
				# Check parameters
				beh.parameters.size.times do |i| 
					if beh.parameters[i].eql?(from)
					
						#puts "Alpha AgentRef: " + from.to_s + " --> " + to.to_s + " in " + beh.agent.to_s
						
						beh.parameters[i]=to												
					 
					end
					
				end 
				
			when Agent
			
				raise "ERROR IN ALPHACONVERT: AGENT FOUND " + beh.to_s
	
			end
		
	end
	
	# Cleans up empty compositions
	def cleanUp(node)
		
		# Test, if composition
		if !node.value === Composition  
			return 
			end
		
		# Test if empty
		if node.value.parts.size > 0 
			return
			end
			
		
		# Iterate until a parent composition with more than one entry is found
		while true
		
			#puts "CLEANING NODE TYPE " + node.value.class.to_s + " in = " + node.parent.value.class.to_s

			p = node.parent
			
			if !p.value
				return
				end
			
			if p.value === Agent
			
				#puts "AGENT"
			
				p.value.behavior = nil
				return
				end
			if  p.value.parts.size > 1
				# Remove sub-tree at high level composition
				
				index = p.value.parts.index(node.value)
				
				#puts "DELETING at " + index.to_s

				p.value.parts.delete_at(index) if index
				return
				end
			node = p
		end

	end
		
	#
	# Optimizes the restriction list of an agent by applying SC-RES-INACT
	#
	def optimizeRestrictions(agent)
	
		#puts agent.to_definition
	
		newRestrictions = Array.new
		agent.restrictions.flatten.each do |r|
		
			# Keep restriction only if found somewhere
			if testName(agent.behavior,r)
				newRestrictions << r
				end
		end
	
		#puts newRestrictions.inspect
		
		agent.restrictions=newRestrictions
	end
	
	#
	# Test if a given name is contained in an expression
	#
	def testName(beh,r)
	
		result = false
	
		case beh
		
			when Composition, Summation:
			
				beh.parts.each do |p| 			
					result = result || testName(p,r) 
				end		
				
			when Input:
				result = beh.object.eql?(r)
				beh.subjects.each do |s|
					result = result || s.eql?(s)
				end if beh.subjects
				result = result || testName(beh.successor,r) if beh.successor
			
			when Output:
				result = beh.object.eql?(r)
				beh.subjects.each do |s|
					result = result || s.eql?(s)
				end if beh.subjects
			
			when AgentRef:				
				result = result || beh.parameters.include?(r) if beh.parameters			
				
			when Tau:
				result = testName(beh.successor,r) if beh.successor
		
		end
		
		return result
	
	end
	
	#
	# Generates a "deep-copy" of the behavior until subsequent agent definitions are found
	#
	def flatCopy(beh, *bool)
	
		rootLevel = true
	
		if bool.size > 0
			rootLevel = bool[0]
			end
		
		case beh
		
			when Input:
				inp = beh.clone
				inp.successor = flatCopy(beh.successor,false) if beh.successor
				return inp
				
			when Output:
				out = beh.clone
				# Not required for async pi, but kept for re-usability
				out.successor = flatCopy(beh.successor,false) if beh.successor		
				return out
				
			when Tau:
				tau = beh.clone
				tau.successor = flatCopy(beh.successor,false) if beh.successor
				return tau
				
			when Composition:
				comp = Composition.new
				comp.parts = beh.parts.collect do |p| flatCopy(p,rootLevel); end
				return comp
				
			when Summation:
				sum = Summation.new
				sum.parts = beh.parts.collect do |p| flatCopy(p,rootLevel); end					
				return sum
				
			when Restriction:			
				# Check if successor exists
				return nil if !beh.successor 				
				# flatCopy Successor
				succ = flatCopy(beh.successor,rootLevel)	
				# Increase restriction counter				
				@restrictionCounter[beh.name] = @restrictionCounter[beh.name] + 1		
				# Look-up current id of name
				new_name = beh.name + "#" + (@restrictionCounter[beh.name] ).to_s
				# Alpha-Convert in result
				alphaConvert(succ,beh.name,new_name)				
				# Add to restriction set
				@restrictions << new_name				
				# Return successor
				return succ
			
			when AgentRef:
				# Find current nesting level to decide if AgentRef should be expanded(!)
				
				# If so...
				if rootLevel	
					# flatCopy Agent of AgentRef
					copy = flatCopy(beh.agent.behavior,rootLevel)
					# Substitute names in flatCopy
					beh.parameters.size.times do |i|
						if beh.parameters[i] != beh.agent.params[i]
							alphaConvert(copy,beh.agent.params[i],beh.parameters[i])
							end
						end if beh.parameters
					# Return flatCopy
					return copy
				else
					# else Return new AgentRe
					return AgentRef.new(beh.agent, beh.parameters.clone)
				end 
			
			else
				raise "Unsupported node type " + beh.class.to_s + " = " + beh.to_s
			
		end
		
	end
	
end

class Node

	attr_accessor :parent, :value
	
	def initialize(parent, value)
		@parent = parent
		@value = value
	end
	
	def to_s
		return value.to_s
	end

end

class IntraactionNodes

	attr_accessor :input, :output
	
	def initialize(output, input)
		@input = input
		@output = output
	end
	
	def to_s
		return output.to_s + " / " + input.to_s 
	end

end

